#!/usr/bin/env python
'''
Created on 20.12.2013

@author: Martin Guenther <mguenthe@uos.de>
'''

import roslib
roslib.load_manifest('race_tabletop_object_publisher')
import rospy

from tabletop_object_detector.srv import TabletopDetection, TabletopDetectionRequest, TabletopDetectionResponse
from tabletop_object_detector.msg import TabletopDetectionResult


class TabletopObjectPublisher:

    def __init__(self):
        rospy.init_node('tabletop_object_publisher')

        # subscribe to object_detection service
        rospy.loginfo('tabletop_object_publisher: waiting for service object_detection...')
        rospy.wait_for_service('object_detection')
        self.object_detection = rospy.ServiceProxy('object_detection', TabletopDetection)
        rospy.loginfo('tabletop_object_publisher: done.')

        # advertise wrapped_object_detection service
        rospy.Service('wrapped_object_detection', TabletopDetection, self.wrapped_object_detection_cb)

        # publisher for tabletop_objects topic
        self.tabletop_objects_pub = rospy.Publisher('tabletop_objects', TabletopDetectionResult)

        rospy.loginfo('tabletop_object_publisher initialized.')

    def wrapped_object_detection_cb(self, req):
        resp = self.object_detection(req)
        self.tabletop_objects_pub.publish(resp.detection)
        return resp

def main():
    TabletopObjectPublisher()
    rospy.spin()

if __name__ == "__main__":
    try:
        main()
    except rospy.ROSInterruptException:
        pass
